import typing
from typing import Generic, Optional, Type, TypeVar

__all__ = ["Convertible", "From"]

T = TypeVar("T")
S = TypeVar("S")
D = TypeVar("D")

_into_reg = {}
_from_reg = {}


class Convertible:
    def __init_subclass__(cls):
        _from_reg[(cls, dict)] = default_struct_from_dict_factory(cls)
        _into_reg[(cls, dict)] = DefaultStructIntoDict

    @classmethod
    def try_from(cls: Type[T], value: S, type_: Optional[Type[S]] = None) -> T:
        converter = _from_reg[(cls, type_ or type(value))]
        return converter().try_from(value)

    def try_into(self, target: Type[T]) -> T:
        converter = _into_reg[(type(self), target)]
        return converter().try_from(self)


class From(Generic[S, D]):
    _src = None
    _dest = None

    @typing._tp_cache
    def __class_getitem__(cls, params):
        src, dest = params

        class Proxy(cls):
            _src = src
            _dest = dest

        return Proxy

    def __init_subclass__(cls):
        cls.register_converter(cls)

    @classmethod
    def try_from(cls, value):
        converter = _from_reg[(cls._dest, cls._src)]
        return converter().try_from(value)

    @classmethod
    def register_converter(cls, conv):
        src, dest = cls._src, cls._dest

        if isinstance(src, type) and issubclass(src, Convertible):
            _into_reg[(src, dest)] = conv
        # src.register_into_converter(dest, cls)

        # if isinstance(dest, type) and issubclass(dest, Convertible):
        _from_reg[(dest, src)] = conv
        # dest.register_from_converter(src, cls)


def default_struct_from_dict_factory(target):
    class DefaultStructFromDict(From):
        def try_from(self, value: dict) -> T:
            return target(**value)

    return DefaultStructFromDict


class DefaultStructIntoDict(From):
    def try_from(self, value: Convertible) -> dict:
        return value.__dict__



import unittest
from dataclasses import dataclass
from typing import NamedTuple, NewType, Annotated, Literal

from convertible import From, Convertible


@dataclass
class User(Convertible):
    age: int
    name: str
    is_active: bool


class _AdminTuple(NamedTuple):
    name: str
    permissions: int


class Admin(Convertible, _AdminTuple):
    pass


# impl From<Test> for Dict


class UserFromDict(From[dict, User]):
    def try_from(self, value: dict) -> User:
        return User(**value)


class UserIntoDict(From[User, dict]):
    def try_from(self, value: User) -> dict:
        return value.__dict__


class UserFromString(From[str, User]):
    def try_from(self, value: str) -> User:
        return User(
            age=23,
            name=value,
            is_active=True,
        )


PublicId = NewType("PublicId", str)
# PublicId = Annotated[str, Literal["PublicId"]]


class UserFromPublicId(From[PublicId, User]):
    def try_from(self, value: PublicId) -> User:
        return User(
            age=32,
            name=value.capitalize(),
            is_active=False,
        )


class AdminFromUser(From[User, Admin]):
    def try_from(self, value: User) -> Admin:
        return Admin(name=value.name, permissions=10011)


class UserFromAdmin(From[Admin, User]):
    def try_from(self, value: Admin) -> User:
        return User(name=value.name, age=22, is_active=True)


class TestConversion(unittest.TestCase):
    def test(self):
        data = {"age": 34, "name": "foo", "is_active": True}
        obj = User.try_from(data)
        assert data == obj.try_into(dict)

    def test_from_string(self):
        obj = User.try_from("foo")
        data = obj.try_into(dict)
        assert obj.is_active == data["is_active"] == True

    def test_from_public_id(self):
        obj = User.try_from("bar", PublicId)
        data = obj.try_into(dict)
        assert obj.is_active == data["is_active"] == False

    def test_cross_conversion(self):
        user_1 = User.try_from("foo")
        admin_1 = user_1.try_into(Admin)
        admin_2 = Admin.try_from(user_1)
        assert admin_1.name == admin_2.name
        user_2 = admin_1.try_into(User)
        assert user_2.name == admin_2.name

    def test_conversion(self):
        user = From[PublicId, User].try_from("bar")
        assert not user.is_active
        admin = From[User, Admin].try_from(user)
        assert admin.name == "Bar"

    def test_manual_register(self):
        class IntFromString:
            def try_from(self, value):
                return int(value)

        From[str, int].register_converter(IntFromString)

        value = From[str, int].try_from("123")
        assert value == 123


if __name__ == "__main__":
    unittest.main()
